import re
from sqlalchemy import (
    create_engine,
    Column,
    ForeignKey,
    Integer,
    String,
)
from sqlalchemy.orm import relationship, sessionmaker, as_declarative, declared_attr

@as_declarative()
class Base(object):
    @declared_attr
    def __tablename__(cls):
        return re.sub(r"(?<!^)(?=[A-Z])", "_", cls.__name__).lower()

class Country(Base):
    id = Column(Integer, primary_key=True)
    name = Column(String(32), nullable=False)
    
    def __repr__(self):
      return f'Country(id={self.id}, name="{self.name}")'
    
    def __str__(self):
      return self.name

class City(Base):
    id = Column(Integer, primary_key=True)
    name = Column(String(32), nullable=False)
    country_id = Column(ForeignKey(Country.id), nullable=False)

    country = relationship(Country)
    
    def __repr__(self):
      return f'City(id={self.id}, name="{self.name}", country_id={self.country_id})'
    
    def __str__(self):
      return self.name

class PostalCode(Base):
    id = Column(Integer, primary_key=True)
    code = Column(String(10), nullable=False)
    city_id = Column(ForeignKey(City.id), nullable=False)

    city = relationship(City)
    
    def __str__(self):
      return f'PostalCode(id={self.id}, code="{self.code}", city_id={self.city_id})'
    
    def __str__(self):
      return self.code

class Address(Base):
    id = Column(Integer, primary_key=True)
    street = Column(String(256), nullable=False)
    postal_code_id = Column(ForeignKey(PostalCode.id), nullable=False)

    postal_code = relationship(PostalCode)
    
    def __str__(self):
      return f'Address(id={self.id}, street="{self.street}", postal_code_id={self.postal_code_id})'
    
    def __str__(self):
      return self.street+", "+str(self.city)+", "+str(self.postal_code)+", "+str(self.country)

    @property
    def city(self):
        return self.postal_code.city

    @property
    def country(self):
        return self.postal_code.city.country
      
engine = create_engine('sqlite://', echo=False)

Session = sessionmaker()
Session.configure(bind=engine)
session = Session()

Base.metadata.create_all(engine)

england = Country(name="England")

london = City(name="London")
london.country = england

westminster = PostalCode(code="SW1A 2AA")
westminster.city = london

downing_street = Address(street="10 Downing Street")
downing_street.postal_code = westminster

session.add(england)
session.add(london)
session.add(westminster)
session.add(downing_street)
session.commit()
